/*
Copyright 2009 Harteex Productions
License: Creative Commons Attribution 3.0 Unported
http://creativecommons.org/licenses/by/3.0/
*/
#include <stdlib.h>
#include <stdio.h>
#include <fsys.h>
#include <string.h>
#include <stdbool.h>
#include <stdint.h>
#include <ucos2.h>
#include <entry.h>

#include "lib/graphics.h"
#include "lib/display.h"
#include "lib/control.h"
#include "lib/timer.h"
#include "lib/fixmath.h"
#include "list.h"
#include "string_ext.h"

#include "mixer.h"
#include "images.h"
#include "list.h"

//#include "debug.h"

#define max(x, y) (x > y ? x : y)
#define min(x, y) (x < y ? x : y)
#define SAFE_FREE(p) { if (p) {free(p); p = 0;}}

#define VERSION_MAJOR    1
#define VERSION_MINOR    0
#define VERSION_REVISION 0

#define TYPE_FOLDER 0
#define TYPE_OGG    1
#define TYPE_XM     2
#define TYPE_MOD    3
#define TYPE_S3M    4

#define SEEK_HOLD_TIME 30

typedef struct {
	uint8_t type;
	char* name;
	void* folderLink; // NULL if it's a file
} fileNode;

typedef struct {
	uint32_t count;
	char path[512];
	fileNode* files;
	void* parentFolder;
	int32_t curIndex;
} folderNode;

list foldersToProcess;

char* gamePathInit(const char* inPath);
bool loadRes();
void unloadRes();
void freeFolder(folderNode* folder);
void scanFolders();
void scanFolder(folderNode* folder);
void scanForFileType(folderNode* folder, char* filter, list* targetList, uint8_t type);
bool startNextSong(bool repeat);
bool startPrevSong();
bool startFile(uint32_t index);
int sortCompare(const void* a, const void* b);
void scrollTo(uint32_t index, uint32_t nbrOfFiles);

gfx_texture* gameFont = NULL;
gfx_texture* gameFontYellow = NULL;
gfx_texture* gameFontWhite = NULL;
gfx_texture* bg = NULL;
gfx_texture* splash = NULL;
gfx_texture* volumeOff = NULL;
gfx_texture* volumeLow = NULL;
gfx_texture* volumeHigh = NULL;
gfx_texture* volumeMeter = NULL;
gfx_texture* repeat = NULL;
gfx_texture* noRepeat = NULL;
gfx_texture* controlPause = NULL;
gfx_texture* controlPlay = NULL;
gfx_texture* controlStop = NULL;
gfx_texture* progressPoint = NULL;
gfx_texture* curPlayingPointer = NULL;
gfx_texture* iconFolder = NULL;
gfx_texture* iconFile = NULL;

display* gameDisplay  = NULL;
timer*   gameTimer    = NULL;
uint32_t gameTickRate = (timer_resolution / 30);
bool     gameRunning  = true;
char     gamePath[256];

bool oggError = false;

folderNode* root = NULL;
folderNode* curFolder = NULL;
folderNode* folderPlaying = NULL;
int32_t curPlaying = 0;
uint32_t fileListOffset = 0;
uint8_t fileListMaxVisible = 12;
bool stopped = true;
uint32_t dpadRightHoldTime = 0;
uint32_t dpadLeftHoldTime = 0;
bool seekStopRight = false;
bool seekStopLeft = false;
bool volChanged = false;

#define PLAYMODE_ONE           0
#define PLAYMODE_ALL           1
#define PLAYMODE_REPEAT_ONE    2
#define PLAYMODE_REPEAT_ALL    3
#define PLAYMODE_RANDOM        4 // NOT USED NOW
#define PLAYMODE_RANDOM_REPEAT 5 // NOT USED NOW
uint8_t playMode = PLAYMODE_REPEAT_ALL;


char* gamePathInit(const char* inPath) {
	uintptr_t i, j;
	for(i = 0, j = 0; inPath[i] != '\0'; i++) {
		if((inPath[i] == '\\') || (inPath[i] == '/'))
			j = i + 1;
	}
	strncpy(gamePath, inPath, j);
	return gamePath;
}

/*void gameScreenshot() {
	char tempString[256];
	unsigned long int tempNumber = 0;
	FILE* tempFile;
	while(true) {
		sprintf(tempString, "%sscreenshot%lu.tga", gamePath, tempNumber);
		tempFile = fsys_fopen(tempString, "rb");
		if(tempFile == NULL)
			break;
		fsys_fclose(tempFile);
		tempNumber++;
	}
	gfx_tex_save_tga(tempString, gfx_render_target);
}*/

bool loadRes()
{
	// splash is loaded in main
	gameFont = gfx_tex_load_tga_from_buffer(tgaFont);
	bg = gfx_tex_load_tga_from_buffer(tgaBg);
	volumeOff = gfx_tex_load_tga_from_buffer(tgaVolumeOff);
	volumeLow = gfx_tex_load_tga_from_buffer(tgaVolumeLow);
	volumeHigh = gfx_tex_load_tga_from_buffer(tgaVolumeHigh);
	volumeMeter = gfx_tex_load_tga_from_buffer(tgaVolumeMeter);
	repeat = gfx_tex_load_tga_from_buffer(tgaRepeat);
	noRepeat = gfx_tex_load_tga_from_buffer(tgaNoRepeat);
	controlPause = gfx_tex_load_tga_from_buffer(tgaControlPause);
	controlPlay = gfx_tex_load_tga_from_buffer(tgaControlPlay);
	controlStop = gfx_tex_load_tga_from_buffer(tgaControlStop);
	progressPoint = gfx_tex_load_tga_from_buffer(tgaProgressPoint);
	curPlayingPointer = gfx_tex_load_tga_from_buffer(tgaCurPlayingPointer);
	iconFolder = gfx_tex_load_tga_from_buffer(tgaFolder);
	iconFile = gfx_tex_load_tga_from_buffer(tgaFile);

	if (gameFont == NULL || bg == NULL || volumeOff == NULL || volumeLow == NULL ||
		volumeHigh == NULL || volumeMeter == NULL || repeat == NULL || noRepeat == NULL ||
		controlPause == NULL || controlPlay == NULL || controlStop == NULL ||
		progressPoint == NULL || curPlayingPointer == NULL || iconFolder == NULL ||
		iconFile == NULL || splash == NULL)
	{
		return false;
	}

	gameFontYellow = gfx_tex_allocate(gameFont->width, gameFont->height);
	gameFontWhite = gfx_tex_allocate(gameFont->width, gameFont->height);

	if (gameFontYellow == NULL || gameFontWhite == NULL)
		return false;

	uint16_t* tempSrcTexPtr = gameFont->address;
	uint16_t* tempTrgTexPtrYellow = gameFontYellow->address;
	uint16_t* tempTrgTexPtrWhite = gameFontWhite->address;
	uintptr_t ptr;
	for (ptr = 0; ptr < (gameFont->width * gameFont->height); ptr++)
	{
		if (tempSrcTexPtr[ptr] == 0xF81F)
		{
			// Preserve transparency
			tempTrgTexPtrYellow[ptr] = tempSrcTexPtr[ptr];
			tempTrgTexPtrWhite[ptr] = tempSrcTexPtr[ptr];
		}
		else
		{
			tempTrgTexPtrYellow[ptr] = gfx_color_rgb(0xFF, 0xFF, 0x00);
			tempTrgTexPtrWhite[ptr] = gfx_color_rgb(0xFF, 0xFF, 0xFF);
		}
	}

	return true;
}

void unloadRes()
{
	freeFolder(root);

	SAFE_FREE(gameFont);
	SAFE_FREE(gameFontYellow);
	SAFE_FREE(gameFontWhite);
	SAFE_FREE(bg);
	SAFE_FREE(splash);
	SAFE_FREE(volumeOff);
	SAFE_FREE(volumeLow);
	SAFE_FREE(volumeHigh);
	SAFE_FREE(volumeMeter);
	SAFE_FREE(repeat);
	SAFE_FREE(noRepeat);
	SAFE_FREE(controlPause);
	SAFE_FREE(controlPlay);
	SAFE_FREE(controlStop);
	SAFE_FREE(progressPoint);
	SAFE_FREE(curPlayingPointer);
	SAFE_FREE(iconFolder);
	SAFE_FREE(iconFile);
}

void freeFolder(folderNode* folder)
{
	if (folder != NULL)
	{
		if (folder->files != NULL)
		{
			uint32_t i;
			for (i = 0; i < folder->count; i++)
			{
				if (folder->files[i].type == TYPE_FOLDER && strcmp(folder->files[i].name, "..") != 0)
				{
					freeFolder(folder->files[i].folderLink);
				}

				SAFE_FREE(folder->files[i].name);
			}

			free(folder->files);
		}

		free(folder);
	}
}

uint8_t _rand8()
{
	uint32_t tempRand = rand();
	return ((tempRand >> 24) ^ (tempRand >> 16) ^ (tempRand >> 8) ^ tempRand);
}

void scanFolders()
{
	foldersToProcess.count = 0;
	foldersToProcess.first = NULL;

	root = (folderNode*)malloc(sizeof(folderNode));
	curFolder = root;
	folderPlaying = curFolder;
	sprintf(root->path, "%s", gamePath);
	root->curIndex = 0;
	root->parentFolder = NULL;
	listInsertLast(&foldersToProcess, root);

	while (foldersToProcess.count > 0)
	{
		listNode* ln = (listNode*)listGetFirstNode(&foldersToProcess);
		folderNode* fn = (folderNode*)ln->data;
		ln->data = NULL;
		listDeleteFirst(&foldersToProcess);

		scanFolder(fn);
	}
}

void scanFolder(folderNode* folder)
{
	char tempString[256];

	// SCAN FOLDER
	list tempList;
	tempList.count = 0;
	tempList.first = NULL;
	findData fd;

	if (folder->parentFolder != NULL)
	{
		fileNode* data = (fileNode*)malloc(sizeof(fileNode));
		char* str = (char*)malloc(3);
		str[0] = '.';
		str[1] = '.';
		str[2] = 0;
		data->type = TYPE_FOLDER;
		data->name = str;
		data->folderLink = folder->parentFolder;

		listInsertLast(&tempList, data);
	}

	// Scan for folders
	fd.handle = 0;
	sprintf(tempString, "%s*", folder->path);
	if (fsys_findfirst(tempString, 0x10, &fd) == 0)
	{
		do
		{
			int length = strlen(fd.filename);
			if (length > 544)
				length = 544;

			fileNode* data = (fileNode*)malloc(sizeof(fileNode));
			char* str = (char*)malloc(length + 1); // We want \0 as well...
			memcpy(str, fd.filename, length + 1);
			data->type = TYPE_FOLDER;
			data->name = str;

			folderNode* fn = (folderNode*)malloc(sizeof(folderNode));
			fn->parentFolder = folder;
			sprintf(fn->path, "%s%s\\", folder->path, str);
			fn->curIndex = 0;
			
			listInsertLast(&foldersToProcess, fn);

			data->folderLink = fn;

			listInsertLast(&tempList, data);
		}
		while (fsys_findnext(&fd) == 0);
	}

	fsys_findclose(&fd);

	scanForFileType(folder, "*.ogg", &tempList, TYPE_OGG);
	scanForFileType(folder, "*.mod", &tempList, TYPE_MOD);
	scanForFileType(folder, "*.xm", &tempList, TYPE_XM);
	scanForFileType(folder, "*.s3m", &tempList, TYPE_S3M);

	// Move data to an array, easier to handle
	if (tempList.count > 0)
	{
		folder->count = tempList.count;
		folder->files = (fileNode*)malloc(tempList.count * sizeof(fileNode));

		listNode* node = tempList.first;
		uint16_t i = 0;
		do
		{
			fileNode* f = (fileNode*)node->data;
			folder->files[i].type = f->type;
			folder->files[i].name = f->name;
			folder->files[i].folderLink = f->folderLink;
			free(node->data);
			node->data = NULL; // Set it to NULL so that listDeleteAll doesn't try to free the data
			node = node->next;
			i++;
		}
		while (node != NULL);
	}

	listDeleteAll(&tempList);

	qsort(folder->files, folder->count, sizeof(fileNode), sortCompare);
}

void scanForFileType(folderNode* folder, char* filter, list* targetList, uint8_t type)
{
	char tempString[256];
	findData fd;
	fd.handle = 0;

	sprintf(tempString, "%s%s", folder->path, filter);
	if (fsys_findfirst(tempString, 0x0, &fd) == 0)
	{
		do
		{
			int length = strlen(fd.filename);
			if (length > 544)
				length = 544;

			fileNode* data = (fileNode*)malloc(sizeof(fileNode));
			char* str = (char*)malloc(length + 1); // We want \0 as well...
			memcpy(str, fd.filename, length + 1);
			data->type = type;
			data->name = str;
			data->folderLink = NULL;

			listInsertLast(targetList, data);
		}
		while (fsys_findnext(&fd) == 0);
	}

	fsys_findclose(&fd);
}

bool startNextSong(bool repeat)
{
	if (folderPlaying->count == 0)
		return false;

	uint32_t i = curPlaying;

	do
	{
		i++;
		i = i % folderPlaying->count;
		if (i == curPlaying) // We didn't find any file in the folder
		{
			stopped = true;
			return false;
		}
	} while (folderPlaying->files[i].type == TYPE_FOLDER);

	if (i < curPlaying && !repeat)
	{
		stopped = true;
		return false;
	}

	startFile(i);
	return true;
}

bool startPrevSong()
{
	if (folderPlaying->count == 0)
		return false;

	uint32_t i = curPlaying;

	do
	{
		if (i == 0)
			i = folderPlaying->count - 1;
		else
			i--;

		if (i == curPlaying) // We didn't find any file in the folder
		{
			stopped = true;
			return false;
		}
	} while (folderPlaying->files[i].type == TYPE_FOLDER);

	startFile(i);
	return true;
}

bool startFile(uint32_t index)
{
	if (folderPlaying->files[index].type == TYPE_FOLDER)
	{
		stopped = true;
		return false;
	}

	char tempString[512];
	curPlaying = index;
	mixerPause(false);
	sprintf(tempString, "%s%s", folderPlaying->path, folderPlaying->files[curPlaying].name);

	bool p = mixerIsPaused();
	mixerPause(true);

	if (folderPlaying->files[index].type == TYPE_OGG)
	{
		oggError = !mixerSetOgg(tempString);
		stopped = oggError;
	}
	else
	{
		oggError = !mixerSetMod(tempString);
		stopped = oggError;
	}

	mixerPause(p);

	return !oggError;
}

int sortCompare(const void* a, const void* b)
{
	fileNode* f1 = (fileNode*)a;
	fileNode* f2 = (fileNode*)b;
	if ((f1->type == TYPE_FOLDER && f2->type == TYPE_FOLDER) || (f1->type != TYPE_FOLDER && f2->type != TYPE_FOLDER))
	{
		return strcmpi(f1->name, f2->name);
	}
	else if (f1->type == TYPE_FOLDER)
	{
		return -1;
	}
	else
	{
		return 1;
	}
}

void scrollTo(uint32_t index, uint32_t nbrOfFiles)
{
	if (nbrOfFiles <= fileListMaxVisible)
	{
		fileListOffset = 0;
	}

	// Check if selection is already in view
	if (index >= fileListOffset && index < fileListOffset + fileListMaxVisible)
	{
		return;
	}
	else if (index < fileListOffset)
	{
		fileListOffset = index;
	}
	else
	{
		fileListOffset = index - (fileListMaxVisible - 1);
	}
}

int main(int argc, char** argv) {
	gamePathInit(argv[0]);
	int ref = EXIT_SUCCESS;

	srand(OSTimeGet());

	control_init();
	gameDisplay = display_create(320, 240, 320, (DISPLAY_FORMAT_RGB565 | DISPLAY_BUFFER_STATIC), NULL, NULL);
	if(gameDisplay == NULL) {
		control_term();
		return ref;
	}
	gfx_init(gameDisplay);

	splash = gfx_tex_load_tga_from_buffer(tgaSplash);
	if (splash != NULL)
	{
		gfx_tex_draw(0, 0, splash);
		display_flip(gameDisplay);
	}

	control_lock(timer_resolution / 4);

	gameTimer   = timer_create();

	uint32_t tempTick  = 0;
	uint32_t tempPauseOS = 0;

	mixerInit(44100, 16, 2, 20);

	char tempString[256];

	if (loadRes())
	{	
		scanFolders();

		uint16_t i;
		for (i = 0; i < 250; i++)
		{
			OSTimeDly(1);

			control_poll();
			if (control_pressed_once(CONTROL_BUTTON_A) || control_pressed_once(CONTROL_BUTTON_START))
			{
				break;
			}
		}
	}
	else
	{
		gameRunning = false;
	}

	int sysref;
	while(gameRunning) {
		sysref = _sys_judge_event(NULL);
		if(sysref < 0) {
			ref = sysref;
			break;
		}

		tempTick += timer_delta(gameTimer);
		if(tempTick < gameTickRate) {
			tempPauseOS = (((gameTickRate - tempTick) * OS_TICKS_PER_SEC) / timer_resolution);
			if(tempPauseOS > 0) OSTimeDly(tempPauseOS);
			while(tempTick < gameTickRate)
				tempTick += timer_delta(gameTimer);
		}


		if(tempTick > (gameTickRate << 2))
			tempTick = (gameTickRate << 2); // HACK - Fixes timing glitch of unknown origin.

		// --- UPDATE ---
		while (tempTick >= gameTickRate)
		{
			tempTick -= gameTickRate;

			control_poll();
			if (control_check(CONTROL_BUTTON_START).pressed && control_check(CONTROL_BUTTON_SELECT).pressed)
			{
				gameRunning = false;
				break;
			}

			if (control_pressed_once(CONTROL_BUTTON_A))
			{
				if (curFolder->count > 0)
				{
					if (curFolder->files[curFolder->curIndex].type == TYPE_FOLDER)
					{
						folderNode* f = curFolder;
						curFolder = curFolder->files[curFolder->curIndex].folderLink;

						if (curFolder != f->parentFolder)
							curFolder->curIndex = 0;

						scrollTo(curFolder->curIndex, curFolder->count);
					}
					else
					{
						folderPlaying = curFolder;
						startFile(curFolder->curIndex);
					}
				}
			}

			if (control_pressed_once(CONTROL_BUTTON_B))
			{
				if (control_check(CONTROL_POWER).pressed)
				{
					gameRunning = false;
					break;
				}
				else
				{
					//gameScreenshot();

					if (curFolder != root)
					{
						curFolder = curFolder->parentFolder;
						scrollTo(curFolder->curIndex, curFolder->count);
					}
				}
			}

			if (control_pressed_once(CONTROL_BUTTON_X))
			{
				if (mixerIsPaused() && songLength > 0)
				{
					stopped = false;
				}
				mixerPause(!mixerIsPaused());
			}

			if (control_check(CONTROL_BUTTON_Y).pressed)
			{
				if (control_pressed_once(CONTROL_DPAD_UP) || control_pressed_once(CONTROL_DPAD_DOWN))
				{
					volChanged = true;
				}
			}
			else if (control_check(CONTROL_BUTTON_Y).changed)
			{
				if (!volChanged)
				{
					playMode++;
					if (playMode > 3)
						playMode = 0;
				}

				volChanged = false;
			}

			if (control_pressed_once(CONTROL_DPAD_DOWN) && !control_check(CONTROL_POWER).pressed &&
				!control_check(CONTROL_BUTTON_Y).pressed)
			{
				curFolder->curIndex++;
				if (curFolder->curIndex >= curFolder->count)
				{
					curFolder->curIndex = 0;
				}
				
				scrollTo(curFolder->curIndex, curFolder->count);
			}

			if (control_pressed_once(CONTROL_DPAD_UP) && !control_check(CONTROL_POWER).pressed &&
				!control_check(CONTROL_BUTTON_Y).pressed)
			{
				curFolder->curIndex--;
				if (curFolder->curIndex < 0)
				{
					curFolder->curIndex = curFolder->count - 1;
				}

				scrollTo(curFolder->curIndex, curFolder->count);
			}

			if (control_check(CONTROL_DPAD_LEFT).pressed)
			{
				dpadLeftHoldTime++;
				if (dpadLeftHoldTime >= SEEK_HOLD_TIME)
				{
					if (dpadLeftHoldTime % 4 == 0 && !seekStopLeft)
					{
						int32_t seekTo = songCur - 3000;
						if (seekTo < 0)
						{
							seekTo = 0;
							seekStopLeft = true;
						}

						if (!mixerSeek(seekTo))
						{
							seekStopLeft = true;
						}
					}
				}
			}
			else
			{
				if (dpadLeftHoldTime > 0 && dpadLeftHoldTime < SEEK_HOLD_TIME && curFolder->count > 0)
				{
					if (songCur < 5000)
					{
						startPrevSong();
					}
					else
					{
						startFile(curPlaying);
					}
				}

				dpadLeftHoldTime = 0;
				seekStopLeft = false;
			}

			if (control_check(CONTROL_DPAD_RIGHT).pressed)
			{
				dpadRightHoldTime++;
				if (dpadRightHoldTime >= SEEK_HOLD_TIME)
				{
					if (dpadRightHoldTime % 4 == 0 && !seekStopRight)
					{
						int32_t seekTo = songCur + 3000;
						if (seekTo >= songLength)
						{
							seekTo = songLength - 1;
							seekStopRight = true;
						}

						if (!mixerSeek(seekTo))
						{
							seekStopRight = true;
						}
					}
				}
			}
			else
			{
				if (dpadRightHoldTime > 0 && dpadRightHoldTime < SEEK_HOLD_TIME && curFolder->count > 0)
				{
					startNextSong(true);
				}

				dpadRightHoldTime = 0;
				seekStopRight = false;
			}

			if (control_pressed_once(CONTROL_TRIGGER_LEFT) || 
				(control_pressed_once(CONTROL_DPAD_DOWN) && control_check(CONTROL_POWER).pressed) ||
				(control_pressed_once(CONTROL_DPAD_DOWN) && control_check(CONTROL_BUTTON_Y).pressed))
			{
				mixerVolumeDecrease();
			}

			if (control_pressed_once(CONTROL_TRIGGER_RIGHT) ||
				(control_pressed_once(CONTROL_DPAD_UP) && control_check(CONTROL_POWER).pressed) ||
				(control_pressed_once(CONTROL_DPAD_UP) && control_check(CONTROL_BUTTON_Y).pressed))
			{
				mixerVolumeIncrease();
			}

			if (!mixerUpdate()) // EOF
			{
				if (!oggError && !mixerIsPaused() && !stopped)
				{
					// Play next or repeat
					switch (playMode)
					{
						case PLAYMODE_ONE:
							stopped = true;
							break;
						case PLAYMODE_ALL:
							startNextSong(false);
							break;
						case PLAYMODE_REPEAT_ONE:
							startFile(curPlaying);
							break;
						case PLAYMODE_REPEAT_ALL:
							startNextSong(true);
							break;
					}
				}
			}
		}
		// --- DRAW ---
		gfx_tex_draw(0, 0, bg);
		
		// Draw repeat icons
		if (playMode == PLAYMODE_REPEAT_ONE || playMode == PLAYMODE_REPEAT_ALL)
		{
			gfx_tex_draw_trans(272, 16, repeat);
		}
		else if (playMode == PLAYMODE_ONE || playMode == PLAYMODE_ALL)
		{
			gfx_tex_draw_trans(272, 16, noRepeat);
		}

		if (playMode == PLAYMODE_ALL || playMode == PLAYMODE_REPEAT_ALL)
		{
			gfx_font_print(272 + 16, 20, gameFont, "All");
		}
		else
		{
			gfx_font_print(272 + 16, 20, gameFont, "One");
		}

		// Draw play/pause/stop icon
		if (stopped)
		{
			gfx_tex_draw_trans(152, 190, controlStop);
		}
		else if (mixerIsPaused())
		{
			gfx_tex_draw_trans(152, 190, controlPause);
		}
		else
		{
			gfx_tex_draw_trans(152, 190, controlPlay);
		}

		// Draw artist / track tags or error msg
		if (oggError)
		{
			gfx_font_print_center(150, gameFontWhite, mixerLastError());
		}
		else
		{
			gfx_font_print_center(150, gameFontWhite, infoArtist);
			gfx_font_print_center(165, gameFontWhite, infoTrack);
		}

		// Draw volume
		if (mixerGetVolume() == 0)
		{
			gfx_tex_draw_trans(280, 121, volumeOff);
		}
		else if (mixerGetVolume() <= VOLUME_MAX / 2)
		{
			gfx_tex_draw_trans(280, 121, volumeLow);
		}
		else
		{
			gfx_tex_draw_trans(280, 121, volumeHigh);
		}

		gfx_tex_draw(281, 71, volumeMeter);
		uint32_t v = (mixerGetVolume() * 43) / VOLUME_MAX;
		gfx_rect_fill_draw(283, 72 + 44 - v, 9, v, gfx_color_rgb(0x59, 0xDF, 0x00));

		// Draw time
		if (mixerTypePlaying == MIXER_TYPE_MOD)
			sprintf(tempString, "%3u", songCur);
		else
			sprintf(tempString, "%02u:%02u", (songCur / 1000) / 60, (songCur / 1000) % 60);

		gfx_font_print(12, 198, gameFontWhite, tempString);

		if (mixerTypePlaying == MIXER_TYPE_MOD)
			sprintf(tempString, "%3u", songLength);
		else
			sprintf(tempString, "%02u:%02u", (songLength / 1000) / 60, (songLength / 1000) % 60);

		gfx_font_print(280, 198, gameFontWhite, tempString);

		// Draw progress point
		// start 27, end 294 - 8 = 286 (distance = 259)
		uint32_t progressX;
		if (songLength > 0)
			progressX = (songCur * 259) / songLength;
		else
			progressX = 0;
		
		gfx_tex_draw_trans(progressX + 27, 216, progressPoint);

		// Draw files
		if (curFolder->count == 0)
		{
			gfx_font_print(16, 16, gameFont, "No files found!");
		}
		else
		{
			uint16_t i;
			for (i = 0; i < curFolder->count; i++)
			{
				if (i - fileListOffset >= 0 && i - fileListOffset < fileListMaxVisible)
				{
					char * temp = curFolder->files[i].name;
					if (strlen(curFolder->files[i].name) > 36)
					{
						strncpy(tempString, curFolder->files[i].name, 33);
						tempString[33] = '.';
						tempString[34] = '.';
						tempString[35] = '.';
						tempString[36] = 0;
						temp = tempString;
					}

					if (curFolder->files[i].type == TYPE_FOLDER)
					{
						gfx_tex_draw_trans(14, 16 + (i - fileListOffset) * 10, iconFolder);
					}
					else
					{
						gfx_tex_draw_trans(14, 16 + (i - fileListOffset) * 10, iconFile);
					}

					if (i == curFolder->curIndex)
						gfx_font_print(16 + 12, 16 + (i - fileListOffset) * 10, gameFontYellow, temp);
					else
						gfx_font_print(16 + 12, 16 + (i - fileListOffset) * 10, gameFont, temp);
				}
			}
		}

		// Draw arrow to show currently playing
		if (!stopped && folderPlaying == curFolder)
		{
			int32_t i = curPlaying - fileListOffset;
			if (i >= 0 && i < fileListMaxVisible)
			{
				gfx_tex_draw_trans(3, 16 + i * 10, curPlayingPointer);
			}
		}

		display_flip(gameDisplay);
	}

	unloadRes();

	mixerPause(true);
	mixerClose();

	timer_delete(gameTimer);

	gfx_term();
	display_delete(gameDisplay);
	control_term();

	return ref;
}
