/*
Copyright 2009 Harteex Productions
License: Creative Commons Attribution 3.0 Unported
http://creativecommons.org/licenses/by/3.0/
*/

#ifndef __mixer_h__
#define __mixer_h__

#define min(x, y) (x < y ? x : y)

#include <stdlib.h>
#include <stdint.h>
#include <stdbool.h>
#include <ucos2.h>
#include <string.h>
#include <audio.h>
#include <fsys.h>
#include <ctype.h>

#include "string_ext.h"

#include "libOgg/ivorbisfile.h"
#include "libOgg/ivorbiscodec.h"
#include "libModplay/modplay.h"

#include "config.h"

//#include "debug.h"

#ifdef GEMEI760
#define VOLUME_MAX 100
#define VOLUME_STEP 5
#else
#define VOLUME_MAX 30
#define VOLUME_STEP 1
#endif

char infoArtist[256] = {0};
char infoAlbum[256] = {0};
char infoTrack[256] = {0};

WaveoutHandle woHandle;

#define  TASK_START_PRIO 5
#define  TASK_STK_SIZE   1024                   // Size of each task's stacks (# of WORDs)
#define  NO_TASKS        1                      // Number of identical tasks 

OS_STK   TaskStk[NO_TASKS][TASK_STK_SIZE];      // Tasks stacks 
OS_STK   TaskStartStk[TASK_STK_SIZE];
char     TaskData[NO_TASKS];                    // Parameters to pass to each task

#define MAXFRAGSIZE 4096
//#define MAXFRAGSIZE 1024
#define SOUNDBUFFERS 4

volatile int curReadBuffer;
int curWriteBuffer;
static volatile int soundThreadPause;
static volatile int soundThreadExit;

#define MIXER_TYPE_NONE 0
#define MIXER_TYPE_OGG  1
#define MIXER_TYPE_MOD  2

bool mixerMusicEOF = true;
uint32_t songLength = 0;
uint32_t songCur = 0;

FSYS_FILE* oggFileHandle;
OggVorbis_File* oggFile = NULL;
int oggCurrentSection;
bool oggEOF = true;
char oggBuffer[MAXFRAGSIZE];
int oggBufferUnWritten = 0;
int oggBufferLastWriteLength = 0;
vorbis_info* oggInfo;
uint8_t mixerTypePlaying = MIXER_TYPE_NONE;

MODFILE* modFile = NULL;
bool modLoaded = false;

int8_t mixerVolume = 20;

char _lastError[256] = {0};

void mixerInit(unsigned int samplerate, unsigned short bits, unsigned char channels, unsigned char volume);
void mixerClose();
void mixerPause(bool b);
bool mixerIsPaused();
void mixerVolumeIncrease();
void mixerVolumeDecrease();
int8_t mixerGetVolume();
void mixerSetVolume(int8_t v);
void mixerSetSampleRate(uint16_t rate);
void mixerSetChannels(uint8_t channels);
void mixerSetBits(uint8_t bits);
void mixerStopAllSound();
void _mixerPlay(void *none);
void mixerSetSound(void* buffer, unsigned int size, bool repeat);
void mixerOggClear();
void mixerModClear();
bool mixerSetOgg(char* file);
void _parseOggComment(char* str, int length);
bool mixerSetMod(char* file);
bool mixerUpdate();
void _mixerUpdateOgg();
void _mixerUpdateMod();
char* mixerLastError();
bool mixerSeek(int32_t ms);

uint16_t dataLength[SOUNDBUFFERS];

static unsigned char mixerSoundBufferTotal[SOUNDBUFFERS * MAXFRAGSIZE];  // Sound buffer
void *mixerSoundBuffers[SOUNDBUFFERS] = {                                // Sound buffers
   (void *)(mixerSoundBufferTotal + ((MAXFRAGSIZE * 0))),
   (void *)(mixerSoundBufferTotal + ((MAXFRAGSIZE * 1))),
   (void *)(mixerSoundBufferTotal + ((MAXFRAGSIZE * 2))),
   (void *)(mixerSoundBufferTotal + ((MAXFRAGSIZE * 3)))};


#endif
